//-----------------------------------------------------------------------
//  Java Indexed Serialization Package
//  ----------------------------------
//  JispDemo.java
//-----------------------------------------------------------------------
//
//  COPYRIGHT NOTICE, DISCLAIMER, and LICENSE:
//  
//  If you modify this file, you may insert additional notices
//  immediately following this sentence.
//  
//  Copyright 1996-2003 Scott Robert Ladd.
//  All rights reserved, except as noted herein.
//  
//  This computer program source file is supplied "AS IS". Scott Robert
//  Ladd (hereinafter referred to as "Author") disclaims all warranties,
//  expressed or implied, including, without limitation, the warranties
//  of merchantability and of fitness for any purpose. The Author
//  assumes no liability for direct, indirect, incidental, special,
//  exemplary, or consequential damages, which may result from the use
//  of this software, even if advised of the possibility of such damage.
//  
//  The Author hereby grants anyone permission to use, copy, modify, and
//  distribute this source code, or portions hereof, for any purpose,
//  without fee, subject to the following restrictions:
//  
//      1. The origin of this source code must not be misrepresented.
//  
//      2. Altered versions must be plainly marked as such and must not
//         be misrepresented as being the original source.
//  
//      3. This Copyright notice may not be removed or altered from any
//         source or altered source distribution.
//  
//  The Author specifically permits (without fee) and encourages the use
//  of this source code for entertainment, education, or decoration. If
//  you use this source code in a product, acknowledgment is not required
//  but would be appreciated.
//  
//  Acknowledgement:
//      This license is based on the wonderful simple license that
//      accompanies libpng.
//
//-----------------------------------------------------------------------
//  For more information on this software package, please visit
//  Scott's web site, Coyote Gulch Productions, at:
//
//      http://www.coyotegulch.com
//-----------------------------------------------------------------------

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import javax.swing.*;
import com.coyotegulch.jisp.*;

public class JispDemo extends JPanel
{
    // utility function
    private void centerDialog(JDialog dlg)
    {
        // center the dialog
        Rectangle screenRect = m_frame.getBounds();
        int centerWidth  = screenRect.width  < dlg.getSize().width  ? screenRect.x : screenRect.x + screenRect.width / 2  - dlg.getSize().width / 2;
        int centerHeight = screenRect.height < dlg.getSize().height ? screenRect.y : screenRect.y + screenRect.height / 2 - dlg.getSize().height / 2;
        dlg.setLocation(centerWidth, centerHeight);
    }
    
    // inner classes
    protected static class AppCloser extends WindowAdapter
    {
        public void windowClosing(WindowEvent e)
        {
            System.exit(0);
        }
    }

    // cancel any running threads
    private class ActionCancel extends AbstractAction
    {
        public void actionPerformed(ActionEvent event)
        {
            m_active = false;
        }
    }
    
    // about box
    private class ActionAbout extends AbstractAction
    {
        public void actionPerformed(ActionEvent event)
        {
            AboutJispDialog dialog = new AboutJispDialog(m_frame);

            // center the dialog
            centerDialog(dialog);
            dialog.setVisible(true);
        }
    }

    // build a database, in a separate thread    
    private static String s_nameInput     = "zipcodes.txt";
    private static String s_nameDatabase  = "zip.data";
    private static String s_nameCodeIndex = "zip.index";

    private class BuildThread extends Thread
    {
        public void run()
        {
            // now we're running                
            m_active = true;
        
            try
            {
                // disable UI while working
                setUIEnabled(false);
                
                // clear output display
                m_output.setText("");
                
                // get order and bucket size
                int order   = Integer.valueOf(m_txtOrder.getText()).intValue();
                
                if (order < 5)
                {
                    order = 5;
                    m_txtOrder.setText("5");
                }
                    
                addMessage("Creating ZIP database from " + s_nameInput);
                
                // delete old files
                File killit = new File(s_nameDatabase);
                
                if (killit.exists())
                    killit.delete();
    
                killit = new File(s_nameCodeIndex);
                
                if (killit.exists())
                    killit.delete();
    
                // create database
                IndexedObjectDatabase database = new IndexedObjectDatabase(s_nameDatabase,true);

                // open indexes and attach them to the database                
                BTreeIndex codeIndex = new BTreeIndex(s_nameCodeIndex,order,new ZipKey(),false);
                database.attachIndex(codeIndex);
                
                // flag to break out of loops if an exception is thrown
                boolean broken = false;
    
                // counter
                long counter = 0;
    
                // open text file
                BufferedReader reader = null;
                reader = new BufferedReader(new FileReader(s_nameInput));
        
                // while we have data to read...
                while (m_active && reader.ready())
                {
                    // read a line
                    String line = reader.readLine();
    
                    if (line.length() < 1)
                    {
                        m_active = false;
                        continue;
                    }
    
                    // count it
                    ++counter;
    
                    m_output.setText("finished line #" + counter);
    
                    int i;
    
                    // extract ZIP
                    String textZIP = "";
    
                    for (i = 0; (line.charAt(i) != ','); ++i)
                        textZIP += line.charAt(i);
    
                    int zip = Integer.valueOf(textZIP).intValue();
    
                    // extract city name
                    ++i;
    
                    String city = line.substring(i);
    
                    // find state
                    String state = FindStateByZip.findState(zip);
    
                    // write the record (we only have one index, but we still need the key array!
                    KeyObject [] keyArray = new KeyObject[1];
                    keyArray[0] = new ZipKey(zip);
    
                    DBRecord record = new DBRecord(city,state,zip);
    
                    database.write(keyArray,record);

                    // we wait
                    Thread.yield();
    
                    // and wait again
                    try
                    {
                        Thread.sleep(20);
                    }
                    catch (InterruptedException e)
                    {
                        // stop running on interruption
                        m_active = false;
                    }
                }
            }
            catch (FileNotFoundException not_found)
            {
                addMessage("ERROR: file " + s_nameInput + " was not found");
                m_active = false;
            }
            catch (Exception ex)
            {
                JOptionPane.showMessageDialog(m_this,"Unable to open or convert those files");
                ex.printStackTrace();
                m_active = false;
            }
            
            // mark thread for disposal
            m_thread = null;  
            
            // re-enable UI while working
            setUIEnabled(true);
            
            addMessage(">>> DONE <<<");
        }
    }
        
    private class ActionBuild extends AbstractAction
    {
        public void actionPerformed(ActionEvent event)
        {
            // create and start thread
            m_thread = new BuildThread();
            m_thread.start();
        }
    }
    
    private class ExportThread extends Thread
    {
        public void run()
        {
            // now we're running                
            m_active = true;
        
            try
            {
                // open an output text file
                String filename = m_txtExport.getText();
                
                // if there is no file name, use a default
                if (filename.length() < 1)
                {
                    filename = new String("zipexport.csv");
                    m_txtExport.setText(filename);
                }
                
                PrintWriter csvFile = new PrintWriter(new FileOutputStream(filename));
                
                // disable UI while working
                setUIEnabled(false);

                // clear output display
                m_output.setText("");

                // open database                
                IndexedObjectDatabase database = new IndexedObjectDatabase(s_nameDatabase,false);
                BTreeIndex codeIndex = new BTreeIndex(s_nameCodeIndex);
                database.attachIndex(codeIndex);
                
                // create an iterator
                BTreeIterator iterator = new BTreeIterator(codeIndex);
                
                // while we have data to read...
                while (m_active)
                {
                    DBRecord rec = (DBRecord)database.read(iterator);
                    
                    if (rec != null)
                    {
                        ZipKey key = (ZipKey)iterator.getKey();
                        csvFile.println(key.m_key + "," + rec.m_city + "," + rec.m_state);
                        csvFile.flush();
                        m_output.setText("working on ZIP " + key.m_key);
                    }
                    
                    m_active = iterator.moveNext();

                    // we wait
                    Thread.yield();
    
                    // and wait again
                    try
                    {
                        Thread.sleep(20);
                    }
                    catch (InterruptedException e)
                    {
                        // stop running on interruption
                        m_active = false;
                    }
                }
            }
            catch (Exception ex)
            {
                addMessage("Something broke");
                ex.printStackTrace();
                m_active = false;
            }
            
            // mark thread for disposal
            m_thread = null;  
            
            // re-enable UI while working
            setUIEnabled(true);
            
            addMessage(">>> DONE <<<");
        }
    }
        
    private class ActionExport extends AbstractAction
    {
        public void actionPerformed(ActionEvent event)
        {
            // create and start thread
            m_thread = new ExportThread();
            m_thread.start();
        }
    }
    
    private class ShowThread extends Thread
    {
        public void run()
        {
            // now we're running                
            m_active = true;
        
            try
            {
                // get from and to keys
                int from = Integer.valueOf(m_txtShowFrom.getText()).intValue();
                int to   = Integer.valueOf(m_txtShowTo.getText()).intValue();
                
                if (from <     0) from = 0;
                if (from > 99999) from = 99999;
                if (to   <     0) to   = 0;
                if (to   > 99999) to   = 99999;
                
                if (from > to)
                {
                    int temp = from;
                    from = to;
                    to = temp;
                }

                m_txtShowFrom.setText("" + from);               
                m_txtShowTo.setText("" + to);               
                
                // disable UI while working
                setUIEnabled(false);
                
                // open database                
                IndexedObjectDatabase database = new IndexedObjectDatabase(s_nameDatabase,false);
                BTreeIndex codeIndex = new BTreeIndex(s_nameCodeIndex);
                database.attachIndex(codeIndex);
                
                //PrintStream indexDump = new PrintStream(new FileOutputStream("index.dump"));
                //codeIndex.dumpTree(indexDump);

                // FORWARD                 
                m_output.setText("Showing from " + from + " to " + to);
                
                // create an iterator and set it to the first object
                BTreeIterator iterator = new BTreeIterator(codeIndex);
                
                if (!iterator.moveTo(new ZipKey(from),true))
                {
                    addMessageEnd("'From' ZIP not found.");
                    m_active = false;
                }
                
                // while we have data to read...
                while (m_active)
                {
                    DBRecord rec = (DBRecord)database.read(iterator);
                    
                    if (rec != null)
                    {
                        if (rec.m_zip <= to)
                            addMessageEnd("found: " + rec.m_zip + "," + rec.m_city + "," + rec.m_state);
                        
                        if (rec.m_zip >= to)
                        {
                            m_active = false;
                            continue;
                        }
                    }
                    
                    m_active = iterator.moveNext();

                    // we wait
                    Thread.yield();
    
                    // and wait again
                    try
                    {
                        Thread.sleep(20);
                    }
                    catch (InterruptedException e)
                    {
                        // stop running on interruption
                        m_active = false;
                    }
                }

                // BACKWARD                 
                addMessageEnd("Showing from " + to + " to " + from);
                
                // create an iterator and set it to the last object
                // iterator = new BTreeIterator(codeIndex);
                m_active = true;
                
                if (!iterator.moveTo(new ZipKey(to),true))
                {
                    addMessageEnd("'To' ZIP not found.");
                    m_active = false;
                }
                
                // while we have data to read...
                while (m_active)
                {
                    DBRecord rec = (DBRecord)database.read(iterator);
                    
                    if (rec != null)
                    {
                        if (rec.m_zip >= from)
                            addMessageEnd("found: " + rec.m_zip + "," + rec.m_city + "," + rec.m_state);
                        
                        if (rec.m_zip <= from)
                        {
                            m_active = false;
                            continue;
                        }
                    }
                    
                    m_active = iterator.movePrevious();

                    // we wait
                    Thread.yield();
    
                    // and wait again
                    try
                    {
                        Thread.sleep(20);
                    }
                    catch (InterruptedException e)
                    {
                        // stop running on interruption
                        m_active = false;
                    }
                }
            }
            catch (Exception ex)
            {
                addMessageEnd("Something broke");
                ex.printStackTrace();
                m_active = false;
            }
            
            // mark thread for disposal
            m_thread = null;  
            
            // re-enable UI while working
            setUIEnabled(true);
            addMessageEnd(">>> DONE <<<");
        }
    }
        
    private class ActionShow extends AbstractAction
    {
        public void actionPerformed(ActionEvent event)
        {
            // create and start thread
            m_thread = new ShowThread();
            m_thread.start();
        }
    }
    
    private class ActionZipSearch extends AbstractAction
    {
        public void actionPerformed(ActionEvent event)
        {
            // disable UI while working
            setUIEnabled(false);
            
            // clear output display
            m_output.setText("");
            
            try
            {
                // open database
                IndexedObjectDatabase database = new IndexedObjectDatabase(s_nameDatabase,false);
                BTreeIndex codeIndex = new BTreeIndex(s_nameCodeIndex);
                database.attachIndex(codeIndex);

                // show what we're looking for                
                addMessageEnd("Searching on key: " + m_txtZip.getText());
                
                // Create a key object
                ZipKey key = new ZipKey(Integer.valueOf(m_txtZip.getText()).intValue());
                
                // read object
                DBRecord rec = (DBRecord)database.read(key,codeIndex);
                
                // display record
                addMessageEnd("Found record:\n  city: " + rec.m_city
                                        + "\n  state: " + rec.m_state
                                          + "\n  zip: " + rec.m_zip);
            }
            catch (Exception ex)
            {
                addMessageEnd("Error finding city record from ZIP code");
            }
            
            // re-enable UI while working
            setUIEnabled(true);
            
            addMessageEnd(">>> DONE <<<");
        }
    }
    
    private class ActionDelete extends AbstractAction
    {
        public void actionPerformed(ActionEvent event)
        {
            // disable UI while working
            setUIEnabled(false);
            
            // clear output display
            m_output.setText("");
            
            try
            {
                // open database
                IndexedObjectDatabase database = new IndexedObjectDatabase(s_nameDatabase,false);
                BTreeIndex codeIndex = new BTreeIndex(s_nameCodeIndex);
                database.attachIndex(codeIndex);

                // show what we're looking for                
                addMessageEnd("Deleting key: " + m_txtDelete.getText());
                
                // Create a key object
                ZipKey key = new ZipKey(Integer.valueOf(m_txtDelete.getText()).intValue());
                
                // look up record
                DBRecord rec = (DBRecord)database.read(key,codeIndex);
                
                // did this work?
                if (rec == null)
                    addMessageEnd("No such record found");
                else
                {
                    // build key list; must have keys to both indexes to remove everything
                    KeyObject [] keyArray = new KeyObject[1];
                    keyArray[0] = key;
                    
                    // read object
                    database.remove(keyArray);
                
                    // display record
                    addMessageEnd("Removed record:\n  city: " + rec.m_city
                                              + "\n  state: " + rec.m_state
                                                + "\n  zip: " + rec.m_zip);
                }
            }
            catch (Exception ex)
            {
                addMessageEnd("Error finding city record from ZIP code");
            }
            
            // re-enable UI while working
            setUIEnabled(true);
            
            addMessageEnd(">>> DONE <<<");
        }
    }
    
    // application working storage
    private JFrame      m_frame;    
    private JispDemo    m_this   = this;
    private ClassLoader m_loader;
    private String      m_cwd;

    // threading tools    
    private Thread  m_thread = null;
    private boolean m_active = false;

    // build DB controls
    private JButton    m_btnBuild;
    private JTextField m_txtOrder;

    // export to text file
    private JButton    m_btnExport;
    private JTextField m_txtExport;
    
    // export to text file
    private JButton    m_btnShow;
    private JTextField m_txtShowFrom;
    private JTextField m_txtShowTo;
    
    // ZIP search controls
    private JButton    m_btnZipSearch;
    private JTextField m_txtZip;
    
    // ZIP search controls
    private JButton    m_btnDelete;
    private JTextField m_txtDelete;
    
    // output display area
    private JTextArea  m_output;
    
    // general UI elements
    private JButton    m_btnCancel;
    private JButton    m_btnAbout;
    
    // set UI state
    private void setUIEnabled(boolean enabled)
    {
        m_btnBuild.setEnabled(enabled);
        m_txtOrder.setEnabled(enabled);
        m_btnExport.setEnabled(enabled);
        m_txtExport.setEnabled(enabled);
        m_btnShow.setEnabled(enabled);
        m_txtShowFrom.setEnabled(enabled);
        m_txtShowTo.setEnabled(enabled);
        m_btnZipSearch.setEnabled(enabled);
        m_txtZip.setEnabled(enabled);
        m_btnDelete.setEnabled(enabled);
        m_txtDelete.setEnabled(enabled);
        m_output.setEnabled(enabled);
        m_btnCancel.setEnabled(!enabled);
        m_btnAbout.setEnabled(enabled);
    }
    
    // add a line of text to the output display
    private void addMessage(String text)
    {
        m_output.setText(text + "\n" + m_output.getText());
    }
    
    // add a line of text to the output display
    private void addMessageEnd(String text)
    {
        m_output.setText(m_output.getText() + "\n" + text);
        m_output.setCaretPosition(m_output.getText().length());
    }
    
    // constructor
    public JispDemo(JFrame owner)
    {
        // base class construction
        super();
        
        // remember owning frame
        m_frame = owner;
        
        // get a class loader for loading resources
        m_loader = ResourceAnchor.class.getClassLoader();
        
        // load look-and-feel
        try
        {
            UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
        }
        catch (Exception ex)
        {
            System.err.println("Error loading look-and-feel:\n" + ex);
        }
        
        // get the current directory
        m_cwd = System.getProperty("user.dir");
        
        // set borders and layout for main display
        setLayout(new BorderLayout());

        // build DB controls
        m_btnBuild = new JButton("Build Database");
        m_btnBuild.addActionListener(new ActionBuild());
        m_txtOrder = new JTextField(4);
        m_txtOrder.setText("64");
        
        JPanel buildPanel = new JPanel();
        buildPanel.setLayout(new FlowLayout(FlowLayout.LEFT));
        buildPanel.add(m_btnBuild);
        buildPanel.add(new JLabel(" Order: ",JLabel.RIGHT));
        buildPanel.add(m_txtOrder);
        
        // export controls
        m_btnExport = new JButton("Export to CSV");
        m_btnExport.addActionListener(new ActionExport());
        m_txtExport = new JTextField(20);
        m_txtExport.setText("zipexport.csv");
        
        JPanel exportPanel = new JPanel();
        exportPanel.setLayout(new FlowLayout(FlowLayout.LEFT));
        exportPanel.add(m_btnExport);
        exportPanel.add(new JLabel(" File: ",JLabel.RIGHT));
        exportPanel.add(m_txtExport);
        
        // show controls
        m_btnShow = new JButton("Show");
        m_btnShow.addActionListener(new ActionShow());
        m_txtShowFrom = new JTextField(5);
        m_txtShowFrom.setText("53555");
        m_txtShowTo = new JTextField(5);
        m_txtShowTo.setText("53818");
        
        JPanel showPanel = new JPanel();
        showPanel.setLayout(new FlowLayout(FlowLayout.LEFT));
        showPanel.add(m_btnShow);
        showPanel.add(new JLabel(" From ",JLabel.RIGHT));
        showPanel.add(m_txtShowFrom);
        showPanel.add(new JLabel(" to ",JLabel.RIGHT));
        showPanel.add(m_txtShowTo);

        // search for ZIP
        m_btnZipSearch = new JButton("Find City for ZIP");
        m_btnZipSearch.addActionListener(new ActionZipSearch());
        m_txtZip = new JTextField(5);

        JPanel zipPanel = new JPanel();
        zipPanel.setLayout(new FlowLayout(FlowLayout.LEFT));
        zipPanel.add(m_btnZipSearch);
        zipPanel.add(new JLabel(" ZIP code: ", JLabel.RIGHT));
        zipPanel.add(m_txtZip);
        
        // delete a ZIP code
        m_btnDelete = new JButton("Delete a ZIP Code");
        m_btnDelete.addActionListener(new ActionDelete());
        m_txtDelete = new JTextField(10);

        JPanel deletePanel = new JPanel();
        deletePanel.setLayout(new FlowLayout(FlowLayout.LEFT));
        deletePanel.add(m_btnDelete);
        deletePanel.add(new JLabel(" ZIP code: ", JLabel.RIGHT));
        deletePanel.add(m_txtDelete);

        // various UI elements
        m_btnCancel = new JButton("Cancel");
        m_btnCancel.addActionListener(new ActionCancel());
        m_btnCancel.setEnabled(false);
        m_btnAbout = new JButton("About...");
        m_btnAbout.addActionListener(new ActionAbout());
        
        JPanel cancelPanel = new JPanel();
        cancelPanel.setLayout(new FlowLayout());
        cancelPanel.add(m_btnCancel);
        cancelPanel.add(m_btnAbout);
        
        // create control panel
        JPanel controlPanel = new JPanel();
        controlPanel.setLayout(new VFlowLayout());
        controlPanel.add(buildPanel);
        controlPanel.add(exportPanel);
        controlPanel.add(showPanel);
        controlPanel.add(zipPanel);
        controlPanel.add(deletePanel);
        controlPanel.add(new JLabel(" "));
        controlPanel.add(cancelPanel);
        
        // create output display area
        m_output = new JTextArea();
        m_output.setEditable(false);
        m_output.setBorder(BorderFactory.createEtchedBorder());
        
        JScrollPane outputPanel = new JScrollPane();
        outputPanel.getViewport().add(m_output);

        // add the work panel to the main window
        add(controlPanel,BorderLayout.WEST);
        add(outputPanel,BorderLayout.CENTER);
    }
    
    // main entry point
    public static void main(String[] args)
    {
        try
        {
            String javaVersion = System.getProperty("java.version");
            
            if (javaVersion.compareTo("1.3.0") < 0)
            {
                System.err.println("JispDemo requires a version 1.3.0 or higher JVM");
                System.exit(1);
            }
             
            JFrame frame = new JFrame();
            frame.setTitle(Preferences.getAppName() + " " + Preferences.getAppVersion());
            frame.getContentPane().setLayout(new BorderLayout());
            frame.getContentPane().add(new JispDemo(frame),BorderLayout.CENTER);
            frame.addWindowListener(new AppCloser());
            frame.pack();
            frame.setSize(720,540);
            
            // center the frame
            Rectangle screenRect = frame.getGraphicsConfiguration().getBounds();
            int centerWidth = screenRect.width < frame.getSize().width ? screenRect.x : screenRect.x + screenRect.width/2 - frame.getSize().width/2;
            int centerHeight = screenRect.height < frame.getSize().height ? screenRect.y : screenRect.y + screenRect.height/2 - frame.getSize().height/2;
            frame.setLocation(centerWidth, centerHeight);
            
            frame.setVisible(true);
        }
        catch (Throwable t)
        {
            System.err.println("Uncaught exception: " + t);
            t.printStackTrace();
        }
    }
}
